package maze

import (
	"container/heap"
	"math/rand"
)

/*
We represent a maze as a graph:
- nodes are the "fields" in the graph, we have width*height nodes
- edges between two nodes represent the possibility to walk from node A to B (e.g. if there is a wall between them)

Both nodes and edges can be uniquely identified by an (integer) index, computed like this:
- nodes: index is ascending line-wise, example for 4x2 grid:
         0123
         4567
         That means:
         index := x + width*y
         x, y := index % width, index / width
- edges: index of the (left/top) node times two.
         Last bit is 0 if this edge is to the right, 1 if this edge is to the bottom.
         That means: edge from 3 to 6 is (3*2)+1 = 7 (horizontal wall between 3 and 6).
         index := (x + width*y)*2 + (is_bottom_edge ? 1 : 0)
         x, y, is_bottom_edge := (index/2)%width, index/2/width, index%2
Conversion is done with helper functions below, and sometimes implicit within the solver.
*/

func Clamp(x, min, max int) int {
	if x < min {
		return min
	} else if x > max {
		return max
	} else {
		return x
	}
}

type EdgeInfos struct {
	index                         int
	is_right_edge, is_bottom_edge bool
	x, y                          int
}

// Reconstruct infos about an edge given its edge index
func (self *Maze) IndexToEdgeInfos(index int) EdgeInfos {
	return EdgeInfos{
		index,
		index%2 == 0, index%2 == 1,
		(index / 2) % self.Width, (index / 2) / self.Width,
	}
}

// Get index of an edge, given the left-top node and kind of edge
func (self *Maze) EdgeToIndex(x, y int, is_bottom_edge bool) int {
	z := 0
	if is_bottom_edge {
		z = 1
	}
	return (y*self.Width+x)*2 + z
}

// A maze, with routines to generate and solve mazes
type Maze struct {
	Height, Width  int
	walls          []bool         // true = wall is present
	possibleWalls  []possibleWall // whether a non-present wall could be added without breaking
	entryX, entryY int            // index of entry/exit field - this field is 1 outside of the maze (!)
	exitX, exitY   int
	// solution:
	Solved, solvable             bool  // only if Solved=true, the following fields have meaningful values
	solutionNodes, solutionEdges []int // index list of the optimal solution (traversed nodes/edges)
}

// A wall is possible iff there is an alternative path between the two fields it seperates.
// Only then adding a wall does not make the maze unsolvable.
// If isDirty is set, the values here are outdated and need to be recomputed (lazy computation).
type possibleWall struct {
	isDirty      bool
	isPossible   bool
	possiblePath []int // list of edges that make this wall possible
}

func NewMaze(width, height int) Maze {
	m := Maze{
		height, width,
		make([]bool, width*height*2),
		make([]possibleWall, width*height*2),
		0, 0, 0, 0,
		false, false, nil, nil,
	}
	m.initPossibleWalls()
	return m
}

func NewMazePtr(width, height int) *Maze {
	m := NewMaze(width, height)
	return &m
}

// Compute shortest path between entry and exit using A*
func (self *Maze) Solve() {
	if !self.Solved {
		self.solvable, self.solutionNodes, self.solutionEdges = self.computeShortestPath(
			Clamp(self.entryX, 0, self.Width-1),
			Clamp(self.entryY, 0, self.Height-1),
			Clamp(self.exitX, 0, self.Width-1),
			Clamp(self.exitY, 0, self.Height-1),
			-1)
		self.Solved = true
	}
}

func (self *Maze) CanGoLeft(x, y int) (bool, int) {
	index := self.EdgeToIndex(x-1, y, false)
	return x > 0 && !self.walls[index], index
}
func (self *Maze) CanGoRight(x, y int) (bool, int) {
	index := self.EdgeToIndex(x, y, false)
	return x < self.Width-1 && !self.walls[index], index
}
func (self *Maze) CanGoUp(x, y int) (bool, int) {
	index := self.EdgeToIndex(x, y-1, true)
	return y > 0 && !self.walls[index], index
}
func (self *Maze) CanGoDown(x, y int) (bool, int) {
	index := self.EdgeToIndex(x, y, true)
	return y < self.Height-1 && !self.walls[index], index
}

// Legacy - for testing
func (self *Maze) GenerateSimpleRandom() {
	for i, _ := range self.walls {
		self.walls[i] = rand.Intn(100) < 35
	}
}

// Choose a random entry/exit node with a minimal distance in between.
func (self *Maze) GenerateRandomEntryExit() {
	// i1/i2: number of border fields starting from top-left corner clockwise.
	i1 := rand.Intn(2*self.Width + 2*self.Height)
	// Minimal distance between them: 1/4 of the maze border.
	i2 := (i1 + (self.Width+self.Height)/2 + rand.Intn(self.Width+self.Height)) % (2*self.Width + 2*self.Height)

	if i1 < self.Width {
		self.entryX = i1
		self.entryY = -1
	} else if i1 < self.Width+self.Height {
		self.entryX = self.Width
		self.entryY = i1 - self.Width
	} else if i1 < 2*self.Width+self.Height {
		self.entryX = self.Width - 1 - (i1 - self.Width - self.Height)
		self.entryY = self.Height
	} else {
		self.entryX = -1
		self.entryY = self.Height - 1 - (i1 - 2*self.Width - self.Height)
	}

	if i2 < self.Width {
		self.exitX = i2
		self.exitY = -1
	} else if i2 < self.Width+self.Height {
		self.exitX = self.Width
		self.exitY = i2 - self.Width
	} else if i2 < 2*self.Width+self.Height {
		self.exitX = self.Width - 1 - (i2 - self.Width - self.Height)
		self.exitY = self.Height
	} else {
		self.exitX = -1
		self.exitY = self.Height - 1 - (i2 - 2*self.Width - self.Height)
	}
}

func (self *Maze) initPossibleWalls() {
	// init all possible walls in an empty maze, with the trivial path making them possible
	for y := 0; y < self.Height; y++ {
		for x := 0; x < self.Width; x++ {
			// right edge: (x,y) -> (x+1,y)
			i_right := self.EdgeToIndex(x, y, false)
			if x < self.Width-1 {
				self.possibleWalls[i_right].isPossible = true
				if y < self.Height-1 {
					self.possibleWalls[i_right].possiblePath = []int{
						self.EdgeToIndex(x, y, true),
						self.EdgeToIndex(x, y+1, false),
						self.EdgeToIndex(x+1, y, true),
					}
				} else {
					self.possibleWalls[i_right].possiblePath = []int{
						self.EdgeToIndex(x, y-1, true),
						self.EdgeToIndex(x, y-1, false),
						self.EdgeToIndex(x+1, y-1, true),
					}
				}
			}

			// bottom edge: (x,y) -> (x,y+1)
			i_bottom := self.EdgeToIndex(x, y, true)
			if y < self.Height-1 {
				self.possibleWalls[i_bottom].isPossible = true
				if x < self.Width-1 {
					self.possibleWalls[i_bottom].possiblePath = []int{
						self.EdgeToIndex(x, y, false),
						self.EdgeToIndex(x+1, y, true),
						self.EdgeToIndex(x, y+1, false),
					}
				} else {
					self.possibleWalls[i_bottom].possiblePath = []int{
						self.EdgeToIndex(x-1, y, false),
						self.EdgeToIndex(x-1, y, true),
						self.EdgeToIndex(x-1, y+1, false),
					}
				}
			}
		}
	}
}

func (self *Maze) GenerateValidMaze() {
	// collect possible walls
	candidates := make([]int, 0)
	for index, w := range self.possibleWalls {
		if w.isPossible || w.isDirty {
			candidates = append(candidates, index)
		}
	}

	// draw a wall and set until no possible walls remain
	for len(candidates) > 0 {
		candidateListIndex := rand.Intn(len(candidates))
		candidate := candidates[candidateListIndex]
		// if candidate is dirty - recompute and remove if not
		if self.possibleWalls[candidate].isDirty {
			self.computePossibleWall(candidate)
		}
		// if wall can be set without breaking the maze - set it
		if self.possibleWalls[candidate].isPossible {
			self.walls[candidate] = true
			// all edges from the alternative path might now become required for a solution:
			// we schedule them for recomputation
			for _, idx := range self.possibleWalls[candidate].possiblePath {
				self.possibleWalls[idx].isDirty = true
			}
		}
		// remove from candidate list - either because used or invalid
		candidates[candidateListIndex] = candidates[len(candidates)-1]
		candidates = candidates[:len(candidates)-1]
	}
}

// Check if a wall can possibly be added (iff there is an alternative path between the seperated nodes)
func (self *Maze) computePossibleWall(idx int) {
	infos := self.IndexToEdgeInfos(idx)
	toX := infos.x
	toY := infos.y
	if infos.is_bottom_edge {
		toY += 1
	} else {
		toX += 1
	}
	reachable, _, used_edges := self.computeShortestPath(infos.x, infos.y, toX, toY, idx)
	self.possibleWalls[idx].isDirty = false
	self.possibleWalls[idx].isPossible = reachable
	self.possibleWalls[idx].possiblePath = used_edges
}

// Compute the shortest path between two points, respecting walls, using the A* algorithm.
// Wall with index "ignoreWall" is not used (like it would have been set).
// Output: true if there is a path, index list of visited fields, index list of used edges
func (self *Maze) computeShortestPath(fromX, fromY, toX, toY int, ignoreWall int) (bool, []int, []int) {
	initialEntry := &PriorityQueueEntry{x: fromX, y: fromY, distance: 0}
	queue := NewPriorityQueue(initialEntry, toX, toY)
	queueEntries := make([]*PriorityQueueEntry, self.Width*self.Height)
	queueEntries[initialEntry.ToIndex(self.Width)] = initialEntry
	closed := make([]bool, self.Width*self.Height)
	reached_from := make([]struct{ node, edge int }, self.Width*self.Height) // map node => predecessor node / used edge

	var successors [4]struct{ pos, edge int }

	for queue.Len() > 0 {
		// x = pop first element
		node := heap.Pop(&queue).(*PriorityQueueEntry)
		queueEntries[node.ToIndex(self.Width)] = nil
		if node.x == toX && node.y == toY {
			// terminate - path found
			visitedNodes := make([]int, node.distance+1)
			visitedEdges := make([]int, node.distance)
			index := int(node.ToIndex(self.Width))
			visitedNodes[node.distance] = index
			for i := node.distance - 1; i >= 0; i-- {
				visitedNodes[i] = reached_from[index].node
				visitedEdges[i] = reached_from[index].edge
				index = reached_from[index].node
			}
			return true, visitedNodes, visitedEdges
		}
		closed[node.x+node.y*self.Width] = true
		// get possible successors
		successors[0] = struct{ pos, edge int }{-1, -1}
		successors[1] = struct{ pos, edge int }{-1, -1}
		successors[2] = struct{ pos, edge int }{-1, -1}
		successors[3] = struct{ pos, edge int }{-1, -1}
		if ok, edgeIndex := self.CanGoUp(node.x, node.y); ok && edgeIndex != ignoreWall {
			successors[0] = struct{ pos, edge int }{node.x + (node.y-1)*self.Width, edgeIndex}
		}
		if ok, edgeIndex := self.CanGoDown(node.x, node.y); ok && edgeIndex != ignoreWall {
			successors[1] = struct{ pos, edge int }{node.x + (node.y+1)*self.Width, edgeIndex}
		}
		if ok, edgeIndex := self.CanGoLeft(node.x, node.y); ok && edgeIndex != ignoreWall {
			successors[2] = struct{ pos, edge int }{(node.x - 1) + node.y*self.Width, edgeIndex}
		}
		if ok, edgeIndex := self.CanGoRight(node.x, node.y); ok && edgeIndex != ignoreWall {
			successors[3] = struct{ pos, edge int }{(node.x + 1) + node.y*self.Width, edgeIndex}
		}
		// Handle successors
		for _, successor := range successors {
			index2 := successor.pos
			if index2 < 0 {
				continue
			}
			x2, y2 := index2%self.Width, index2/self.Width
			if !closed[index2] {
				if node2 := queueEntries[index2]; node2 != nil {
					// we found a shorter path to an already explored node
					if node.distance+1 < node2.distance {
						node2.distance = node.distance + 1
						queue.update(node2)
						reached_from[index2] = struct{ node, edge int }{node.ToIndex(self.Width), successor.edge}
					}
				} else {
					// we found a path to a non-explored node
					node2 := &PriorityQueueEntry{x: x2, y: y2, distance: node.distance + 1}
					queueEntries[index2] = node2
					queue.Push(node2)
					reached_from[index2] = struct{ node, edge int }{node.ToIndex(self.Width), successor.edge}
				}
			}
		}
	}
	return false, nil, nil
}
