package maze

import (
	"bytes"
	"encoding/base64"
	"fmt"
	"github.com/disintegration/imaging"
	"image"
	"image/color"
	"image/draw"
	"image/png"
	"os"
)

// Draw line from (x,y) -> (x,y2)
func drawVerticalLine(img *image.RGBA, x, y, y2 int, color color.RGBA, thickness int) {
	x -= thickness / 2
	x2 := x + thickness
	draw.Draw(img, image.Rect(x, y, x2, y2), &image.Uniform{color}, image.Point{}, draw.Src)
}

// Draw line from (x,y) -> (x2,y)
func drawHorizontalLine(img *image.RGBA, x, y, x2 int, color color.RGBA, thickness int) {
	y -= thickness / 2
	y2 := y + thickness
	draw.Draw(img, image.Rect(x, y, x2, y2), &image.Uniform{color}, image.Point{}, draw.Src)
}

// Get an arrow image, rotated to fit entry/exit.
// Rotation = number of 90° counter-clockwise rotations necessary
func drawArrow(img *image.RGBA, x, y int, rotation int) {
	data := `iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAAWhJREFU
SIntk7FLQlEUxr9zfBaBlEM+jAYzNaKG2huafEpLNPQntIYQ0djoJO39AdHm0pAZ0hbU4tIgSiQU
okEWhKH07mkQo8KX8tTt/ZZ7Oed+33e43As4ODgMA//cRkAPxTbtaNmOyHR/eokkrYeMjD8cXxp5
IACAiIjJUKLyvlD0aCqw7h1tYCeXyc3MiTH3eEEPGTvA4b+eAwd+GxHpxHSsh69vfPPRNcsB28u2
S4+83UKozwFkgogWLLttTkw2D16KuSeLwNcmgV39BfY5lpJ3QCUnSUuVSudNYIhX2g1i8gAUr7MK
dmraqMJEqUcS7Nfus6e/huhsphdjM/2asYllArLdekrkA4KU1hhLVipnjb996ibqhS9irDAo/7Mm
IgKRNGm8Vy1kHqy0Q7lSJeqOiBO10sVlr7MDPRoRVVdKdp9nW6u1YqZnGADY+gYeb1AXEGuu1la1
lLtCuazs+Dg4OAAAvgC3n3GqzF8mdgAAAABJRU5ErkJggg==`
	data2, _ := base64.StdEncoding.DecodeString(data)
	arrow, _ := png.Decode(bytes.NewReader(data2))

	rotation = rotation % 4
	if rotation == 0 {
		draw.Draw(img, image.Rect(x, y, x+arrow.Bounds().Max.X, y+arrow.Bounds().Max.Y), arrow, image.Point{}, draw.Over)
	} else if rotation == 1 {
		arrow2 := imaging.Rotate90(arrow)
		draw.Draw(img, image.Rect(x, y, x+arrow2.Bounds().Max.X, y+arrow2.Bounds().Max.Y), arrow2, image.Point{}, draw.Over)
	} else if rotation == 2 {
		arrow2 := imaging.Rotate180(arrow)
		draw.Draw(img, image.Rect(x, y, x+arrow2.Bounds().Max.X, y+arrow2.Bounds().Max.Y), arrow2, image.Point{}, draw.Over)
	} else if rotation == 3 {
		arrow2 := imaging.Rotate270(arrow)
		draw.Draw(img, image.Rect(x, y, x+arrow2.Bounds().Max.X, y+arrow2.Bounds().Max.Y), arrow2, image.Point{}, draw.Over)
	}
}

// Get the rotation of an arrow, given border coordinates.
func (self *Maze) getArrowRotation(x, y, rotationOffset int) int {
	if x < 0 {
		return rotationOffset
	} else if y >= self.Height {
		return (1 + rotationOffset) % 4
	} else if x >= self.Width {
		return (2 + rotationOffset) % 4
	} else {
		return (3 + rotationOffset) % 4
	}
}

// Export maze as image
func (self *Maze) ToImage(scale int, drawSolution bool) *image.RGBA {
	linewidth := 2
	linewidthSolution := 3
	offset := scale + linewidth // (scale - linewidth) / 2
	img := image.NewRGBA(image.Rect(0, 0, self.Width*scale+2*offset, self.Height*scale+2*offset))
	bg := color.RGBA{240, 240, 240, 255}
	fg := color.RGBA{0, 0, 0, 255}
	solutionColor := color.RGBA{200, 0, 0, 255}
	gateColor := bg // color.RGBA{220, 220, 220, 255}
	draw.Draw(img, img.Bounds(), &image.Uniform{bg}, image.Point{}, draw.Src)

	// draw surrounding rectangle
	drawHorizontalLine(img, offset, offset, offset+(self.Width*scale), fg, linewidth)
	drawHorizontalLine(img, offset, offset+(self.Height*scale), offset+(self.Width*scale), fg, linewidth)
	drawVerticalLine(img, offset, offset, offset+(self.Height*scale), fg, linewidth)
	drawVerticalLine(img, offset+(self.Width*scale), offset, offset+(self.Height*scale), fg, linewidth)

	// draw
	for i, is_present := range self.walls {
		if is_present {
			info := self.IndexToEdgeInfos(i)
			if info.is_bottom_edge {
				x := info.x*scale + offset
				y := (info.y+1)*scale + offset
				drawHorizontalLine(img, x, y, x+scale, fg, linewidth)
			} else if info.is_right_edge {
				x := (info.x+1)*scale + offset
				y := info.y*scale + offset
				drawVerticalLine(img, x, y, y+scale, fg, linewidth)
			}
		}
	}

	// solution?
	if drawSolution {
		self.Solve()
		if self.solvable {
			offset2 := offset + scale/2
			for _, edge := range self.solutionEdges {
				edgeInfos := self.IndexToEdgeInfos(edge)
				if edgeInfos.is_bottom_edge {
					drawVerticalLine(img, offset2+(edgeInfos.x*scale), offset2+(edgeInfos.y*scale), offset2+(edgeInfos.y*scale)+scale, solutionColor, linewidthSolution)
				} else {
					drawHorizontalLine(img, offset2+(edgeInfos.x*scale), offset2+(edgeInfos.y*scale), offset2+(edgeInfos.x*scale)+scale, solutionColor, linewidthSolution)
				}
			}
			x, y := self.solutionNodes[0]%self.Width, self.solutionNodes[0]/self.Width
			if x != self.entryX {
				drawHorizontalLine(img, offset2+x*scale, offset2+y*scale, offset2+self.entryX*scale, solutionColor, linewidthSolution)
			} else {
				drawVerticalLine(img, offset2+x*scale, offset2+y*scale, offset2+self.entryY*scale, solutionColor, linewidthSolution)
			}
			x, y = self.solutionNodes[len(self.solutionNodes)-1]%self.Width, self.solutionNodes[len(self.solutionNodes)-1]/self.Width
			if x != self.exitX {
				drawHorizontalLine(img, offset2+x*scale, offset2+y*scale, offset2+self.exitX*scale, solutionColor, linewidthSolution)
			} else {
				drawVerticalLine(img, offset2+x*scale, offset2+y*scale, offset2+self.exitY*scale, solutionColor, linewidthSolution)
			}
		} else {
			_ = fmt.Errorf("Warning: not solvable!\n")
		}
	}

	// draw entry/exit background
	x1 := offset + self.entryX*scale - linewidth/2
	y1 := offset + self.entryY*scale - linewidth/2
	draw.Draw(img, image.Rect(x1, y1, x1+scale+linewidth, y1+scale+linewidth), &image.Uniform{gateColor}, image.Point{}, draw.Src)
	x1 = offset + self.exitX*scale - linewidth/2
	y1 = offset + self.exitY*scale - linewidth/2
	draw.Draw(img, image.Rect(x1, y1, x1+scale+linewidth, y1+scale+linewidth), &image.Uniform{gateColor}, image.Point{}, draw.Src)

	// draw entry/exit arrows
	drawArrow(img, offset+self.entryX*scale, offset+self.entryY*scale, self.getArrowRotation(self.entryX, self.entryY, 0))
	drawArrow(img, offset+self.exitX*scale, offset+self.exitY*scale, self.getArrowRotation(self.exitX, self.exitY, 2))

	return img
}

// Save maze as png image, optionally with solution
func (self *Maze) ToPngImage(scale int, drawSolution bool, filename string) {
	f, err := os.Create(filename)
	if err != nil {
		panic(err)
	}
	err = png.Encode(f, self.ToImage(scale, drawSolution))
	if err != nil {
		panic(err)
	}
}
